#include <Adafruit_SPIFlash.h>
#include <LSM6DS3.h>
#include <nrf52840.h>
#include <Wire.h>

#define DO_WORK_PIN   D2
#define SHUTDOWN_PIN  D3

Adafruit_FlashTransport_QSPI flashTransport;
LSM6DS3 myIMU; // Default is I2C mode

enum commands_t
{
    DA_CMD_DO_NOTHING = 0,
    DA_CMD_BUTTON_PRESS,
    DA_CMD_DOUBLE_TAP,
    DA_CMD_SYSTEM_OFF_SLEEP
};
static int da_command = DA_CMD_DO_NOTHING;
static SemaphoreHandle_t da_xSemaphore;
static int da_work_LED_status = HIGH;

void QSPIF_sleep(void)
{
  flashTransport.begin();
  flashTransport.runCommand(0xB9);
  flashTransport.end();
}

void setupAccelerometer()
{
  // Start with LSM6DS3 in disabled to save power
  myIMU.settings.gyroEnabled = 0;
  myIMU.settings.accelEnabled = 0;

  myIMU.begin();

  /* Values from the application note */

  myIMU.writeRegister(LSM6DS3_ACC_GYRO_CTRL1_XL, 0x60);    // Turn on the accelerometer
                                                           // ODR_XL = 416 Hz, FS_XL = ±2 g
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_TAP_CFG1, 0x8E);    // Enable interrupts and tap detection on X, Y, Z-axis
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_TAP_THS_6D, 0x8C);  // Set tap threshold
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_INT_DUR2, 0x7F);    // Set Duration, Quiet and Shock time windows
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_WAKE_UP_THS, 0x80); // Single & double-tap enabled (SINGLE_DOUBLE_TAP = 1)
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_MD1_CFG, 0x08);     // Double-tap interrupt driven to INT1 pin

  // Set up the sense mechanism to generate the DETECT signal to wake from system_off
  pinMode(PIN_LSM6DS3TR_C_INT1, INPUT_PULLDOWN_SENSE);
  attachInterrupt(digitalPinToInterrupt(PIN_LSM6DS3TR_C_INT1), doubleTapISR, RISING);

  return;
}

void setupWakeUpInterrupt()
{
  // Initially the system kept turning back on after system_off. Through trial and error
  // added the accelerator power down and that fixed the issue
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_MD1_CFG, 0x00);    // Power down the accelerometer
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_CTRL1_XL, 0x00);   // Disable double-tap interrupt

  // Set up the accelerometer for Wake-up interrupt.
  // Per the application note, use a two step set up to avoid spurious interrupts
  // Set up values are from the application note, and then adjusted for minimum power

  myIMU.writeRegister(LSM6DS3_ACC_GYRO_WAKE_UP_DUR, 0x00); // No duration
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_WAKE_UP_THS, 0x02); // Set wake-up threshold
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_TAP_CFG1, 0x80);    // Enable interrupts and apply slope filter; latch mode disabled
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_CTRL1_XL, 0x70);    // Turn on the accelerometer
                                                           // ODR_XL = 833 Hz, FS_XL = ±2 g
  delay(4);                                                // Delay time per application note
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_CTRL1_XL, 0xB0);    // ODR_XL = 1.6 Hz
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_CTRL6_G, 0x10);     // High-performance operating mode disabled for accelerometer
  myIMU.writeRegister(LSM6DS3_ACC_GYRO_MD1_CFG, 0x20);     // Wake-up interrupt driven to INT1 pin

  return;
}

void setup()
{
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT);
  
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  da_work_LED_status = HIGH;
  digitalWrite(LED_BLUE, da_work_LED_status);

  //start bluefruit (this also starts the soft device if you don't start the sd then none of your sd_* calls will do anything
  //Bluefruit.begin();

  QSPIF_sleep();

  setupAccelerometer();

  da_xSemaphore = xSemaphoreCreateBinary();
  da_command = DA_CMD_DO_NOTHING;

  pinMode(DO_WORK_PIN, INPUT_PULLUP_SENSE);
  attachInterrupt(digitalPinToInterrupt(DO_WORK_PIN), doWorkISR, FALLING);

  pinMode(SHUTDOWN_PIN, INPUT_PULLUP_SENSE);
  attachInterrupt(digitalPinToInterrupt(SHUTDOWN_PIN), shutdownISR, FALLING);

  // Flash green to see power on, reset, and wake from system_off
  digitalWrite(LED_GREEN, LOW);
  delay(1000);
  digitalWrite(LED_GREEN, HIGH);;

  // Minimizes power when bluetooth is used
  NRF_POWER->DCDCEN = 1;
}

void doWorkISR()
{
  xSemaphoreGive(da_xSemaphore);
  da_command = DA_CMD_BUTTON_PRESS;
}

void shutdownISR()
{
  da_command = DA_CMD_SYSTEM_OFF_SLEEP;
  xSemaphoreGive(da_xSemaphore);
}

void doubleTapISR()
{
  da_command = DA_CMD_DOUBLE_TAP;
  xSemaphoreGive(da_xSemaphore);
}

void loop()
{
  // FreeRTOS will automatically put the system in system_on sleep mode here
  xSemaphoreTake(da_xSemaphore, portMAX_DELAY);

  // Something caused the system to wake up

  if (da_command == DA_CMD_BUTTON_PRESS)
  {
    // Do work
    da_work_LED_status = !da_work_LED_status;
    digitalWrite(LED_BLUE, da_work_LED_status);
  }
  else if (da_command == DA_CMD_DOUBLE_TAP)
  {
    da_work_LED_status = !da_work_LED_status;
    digitalWrite(LED_BLUE, da_work_LED_status);
    delay(500);
    da_work_LED_status = !da_work_LED_status;
    digitalWrite(LED_BLUE, da_work_LED_status);
  }
  else if (da_command == DA_CMD_SYSTEM_OFF_SLEEP)
  {
    //Flash red to see we are going to system_off sleep mode
    digitalWrite(LED_RED, LOW);
    delay(1000);
    digitalWrite(LED_RED, HIGH);

    // Setup up double tap interrupt to wake back up
    setupWakeUpInterrupt();

    NRF_POWER->SYSTEMOFF=1; // Execution should not go beyond this
    //sd_power_system_off(); // Use this instead if using the soft device
  }
}